@@ -21,8 +21,12 @@ DSN = ""
 SLOT = "example"
 
 
+def event_filter(event):
+    return event['kind'] == "change" and event['action'] == "insert"
+
+
 async def main():
-    consumer = Consumer(DSN, SLOT)
+    consumer = Consumer(DSN, SLOT, event_filter=event_filter)
     await consumer.connect()
 
     while True:
@@ -36,7 +40,15 @@ async def process_transaction(transaction):
     await asyncio.sleep(random.randint(1, 10))
     print(f"Processed transaction {transaction!r}")
 
-    await transaction.flush()
+    transaction.applied()
+
+    # Usually the consumer will send feedback to the server every so
+    # often (15 seconds by default), so if the consumer exits immediately
+    # after marking a transaction as 'applied', the server may not be aware
+    # and feed the same transactions when the consumer restarts.
+    # If potentially repeatedly processing the same transaction has a high
+    # cost manually sending feedback may be worth-while.
+    await transaction.consumer.send_feedback()
 
 
 asyncio.run(main())
