@@ -22,9 +22,10 @@ REPLICATION_STATUS_INTERVAL = 15
 
 
 class Consumer:
-    def __init__(self, dsn, slot_name):
+    def __init__(self, dsn, slot_name, event_filter=None):
         self._dsn = dsn
         self._slot_name = slot_name
+        self._event_filter = event_filter
 
         self._current_transaction_events = None
         self._unflushed_transactions = []
@@ -52,12 +53,17 @@ class Consumer:
         await loop.run_in_executor(None, self._close)
 
     async def read_transaction(self):
-        loop = asyncio.get_running_loop()
-
         while True:
             replication_message = self._cursor.read_message()
 
             if replication_message is None:
+                # If we're not collecting events within a transaction and
+                # there are no unflushed transactions, we can advance
+                # to the most recent (presumably keepalive) message.
+                # This helps with the lag calculations in pg_stat_replication
+                if self._current_transaction_events is None and not self._unflushed_transactions:
+                    self._advance_applied_lsn(self._cursor.wal_end)
+
                 await self._wait_until_readable_or_timeout()
                 continue
 
@@ -66,9 +72,13 @@ class Consumer:
             if self._current_transaction_events is None:
                 if event['kind'] == "message":
                     # Message sent with Transactional=False
-                    txn = Transaction(self, [event], replication_message.data_start, loop.time())
-                    self._unflushed_transactions.append(txn)
-                    return txn
+                    txn = await self._construct_transaction([event], replication_message.data_start)
+
+                    # txn will be None if, after filtering, it contained no
+                    # events (in which case construct_transaction will take care
+                    # of marking it as applied)
+                    if txn:
+                        return txn
                 elif event['kind'] == "begin":
                     self._current_transaction_events = []
                 else:
@@ -76,15 +86,14 @@ class Consumer:
 
             else:
                 if event['kind'] == "commit":
-                    txn = Transaction(self, self._current_transaction_events, replication_message.data_start, loop.time())
+                    txn = await self._construct_transaction(self._current_transaction_events, replication_message.data_start)
                     self._current_transaction_events = None
-                    self._unflushed_transactions.append(txn)
 
-                    # Quietly flush transactions with no events (eg DDL)
-                    if txn.events:
+                    # txn will be None if, after filtering, it contained no
+                    # events (in which case construct_transaction will take care
+                    # of marking it as applied)
+                    if txn:
                         return txn
-                    else:
-                        await txn.flush()
                 elif event['kind'] in ("change", "truncate", "message"):
                     self._current_transaction_events.append(event)
                 else:
@@ -118,10 +127,10 @@ class Consumer:
             else:
                 self._current_batch_transactions.append(txn)
 
-    async def force_feedback(self):
+    async def send_feedback(self):
         loop = asyncio.get_running_loop()
 
-        await loop.run_in_executor(None, self._force_feedback)
+        await loop.run_in_executor(None, self._send_feedback)
 
     def _connect(self):
         self._connection = psycopg2.connect(self._dsn, connection_factory=psycopg2.extras.LogicalReplicationConnection)
@@ -153,7 +162,30 @@ class Consumer:
         loop.remove_reader(fd)
         fut.set_result(None)
 
-    def _force_feedback(self):
+    async def _construct_transaction(self, events, commit_lsn):
+        loop = asyncio.get_running_loop()
+
+        if self._event_filter:
+            events = list(filter(self._event_filter, events))
+
+        # We always generate a Transaction even if we're about to mark it
+        # applied so it can be tracked by unflushed_transactions
+        txn = Transaction(self, events, commit_lsn, loop.time())
+        self._unflushed_transactions.append(txn)
+
+        # Quietly mark the transaction as applied if it's empty
+        if events:
+            return txn
+        else:
+            txn.applied()
+
+    def _advance_applied_lsn(self, lsn):
+        # Update all three LSNs to make pg_stat_replication clearer
+        # This will never block
+        self._cursor.send_feedback(write_lsn=lsn, flush_lsn=lsn, apply_lsn=lsn, force=False)
+
+    def _send_feedback(self):
+        # This might block
         self._cursor.send_feedback(force=True)
 
     def _transaction_became_flushable(self, txn):
@@ -182,11 +214,9 @@ class Consumer:
         earliest_unflushed_txn = self._unflushed_transactions[0]
 
         if earliest_unflushed_txn._flushable:
-            flush_lsn = earliest_unflushed_txn._data_start_lsn
-
             del self._unflushed_transactions[0]
 
-            self._cursor.send_feedback(flush_lsn=flush_lsn, force=False)
+            self._advance_applied_lsn(earliest_unflushed_txn._commit_lsn)
 
 
 class ProtocolError(Exception):
@@ -194,10 +224,10 @@ class ProtocolError(Exception):
 
 
 class Transaction:
-    def __init__(self, consumer, events, data_start_lsn, arrival_time):
-        self._consumer = consumer
+    def __init__(self, consumer, events, commit_lsn, arrival_time):
+        self.consumer = consumer
         self.events = events
-        self._data_start_lsn = data_start_lsn
+        self._commit_lsn = commit_lsn
         self._arrival_time = arrival_time
         self._flushable = False
 
@@ -205,20 +235,15 @@ class Transaction:
         info = [self.__class__.__name__]
 
         info.append(f"events={self.events}")
-        info.append(f"data_start_lsn={lsn_int_to_str(self._data_start_lsn)}")
+        info.append(f"commit_lsn={lsn_int_to_str(self._commit_lsn)}")
         info.append(f"arrival_time={self._arrival_time}")
         info.append(f"flushable={self._flushable}")
 
         return "<{}>".format(" ".join(info))
 
-    async def flush(self):
-        loop = asyncio.get_running_loop()
-
-        await loop.run_in_executor(None, self._flush)
-
-    def _flush(self):
+    def applied(self):
         self._flushable = True
-        self._consumer._transaction_became_flushable(self)
+        self.consumer._transaction_became_flushable(self)
 
 
 def lsn_int_to_str(lsn):
